
/* DBA Videowall Ver 2.0   */
/* A. Schiffler, ZKM, 1999 */

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <signal.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/extensions/XShm.h>
#include <sys/shm.h>
#include <X11/keysym.h>

/* Framegrabber */

#include "ioctl_meteor.h"

/* Setup */

#include "videowall.h"

/* Screen */

#define ROWS 480
#define COLS 640

#define IN_SIZE 2
#define OUT_SIZE 1
#define BYTE_SIZE (ROWS * COLS * OUT_SIZE)
#define RGB_SIZE (ROWS * COLS * IN_SIZE)

#define WINDOW_SIZE_X COLS
#define WINDOW_SIZE_Y ROWS

extern int errno;

/* ============= Customize this ======================== */

/* Framegrabber device */

static char DefaultName[]="/dev/mmetfgrab0";
static char ColortestName[]="/dev/mmetfgrab0";

/* Use multiple X-Servers */

#define SINGLE_SCREEN	0

/* Generate debugging text */

#define BE_VERBOSE	1

/* Colorspace test */

static int colortest[NUM_SCREENS][64*1024];

/* Prototypes */

static void setupDisplay();
void gotframe(int x);

/* Variables */

static Display* display_[NUM_SCREENS];
static GC gc_[NUM_SCREENS];
static Window wtwin_[NUM_SCREENS];

static XImage *shmimage[NUM_SCREENS];
static XShmSegmentInfo* shminfo[NUM_SCREENS];

static char *tbuf[NUM_SCREENS];
static char *bgbuf;

static XColor palette[NUM_SCREENS][256];
static int maxcols = 256;
static Colormap wtcolormap[NUM_SCREENS];

static char *DevName;
static char *mmbuf;
static char *fbuf;

static struct sigaction sigact;

static int framecounter = 0;

static float LastTime;
static int framecounter;

static int num_colors=0;
static int the_colors[255];

static int xstart[NUM_SCREENS],ystart[NUM_SCREENS];
static int xsize[NUM_SCREENS],ysize[NUM_SCREENS];
static int *xblocksize[NUM_SCREENS],*yblocksize[NUM_SCREENS];

static int *xtest[NUM_SCREENS];

/* ============= Routines ============= */

void initColortest()
{
 int i;
 
 for (i=0; i<NUM_SCREENS; i++) {
  memset(colortest[i],0,64*1024*sizeof(int));
 }
}

void loadColortest(char *filename)
{
 int i;
 FILE *infile;

 if ((infile=fopen(filename,"r"))!=NULL) {
  for (i=0; i<NUM_SCREENS; i++) {
   fread(colortest[i],64*1024*sizeof(int),1,infile);
  }
  fclose(infile);
  fprintf(stderr,"Loaded %s ...\n",filename);
 } else {
  fprintf(stderr,"Error loading %s ...\n",filename);
 }
}

void saveColortest(char *filename)
{
 int i;
 FILE *outfile;

 if ((outfile=fopen(filename,"w"))!=NULL) {
  for (i=0; i<NUM_SCREENS; i++) { 
   fwrite(colortest[i],64*1024*sizeof(int),1,outfile);
  }
  fclose(outfile);
  fprintf(stderr,"Saved %s ...\n",filename);
 } else {
  fprintf(stderr,"Error saving %s ...\n",filename);
 }
}

void updateColortest()
{
 int n,x,y;
 unsigned short *curpos;

 /* Process all test areas */ 
 for (n=0; n<NUM_SCREENS; n++) {
  for (y=initial_value[n][0]; y<(initial_value[n][0]+initial_value[n][1]); y++) {
   /* Setup pointer */
   curpos=(unsigned short *)fbuf;
   curpos += initial_value[n][0]+COLS*initial_value[n][2];
   for (x=initial_value[n][2]; x<(initial_value[n][2]+initial_value[n][3]); x++) {
    /* Mark as background */
    colortest[n][*curpos] |= 1;
    /* Advance pointer */
    curpos++;
   }
  }
 }
 fprintf(stderr,"Updated colortest ...\n");

}

/* ------- */

/* Load 640x480 PPM image from XV */
void *LoadImage(char *filename)
{
 int i,j;
 FILE *infile;
 unsigned char *the_data;
 char *the_image;
 char header[80];
 char dummy[80];
 char line[80];
 int xres,yres,cres;
 int cur_color;
 int r,g,b;
 
 if (BE_VERBOSE) fprintf (stderr,"===== Loading PPM Image =====\n\n");

 /* Always return something */
 the_image=(char *)malloc(640*480);
 bzero(the_image,640*480);
 
 infile=fopen(filename,"r");
 fgets(line,80,infile);
 sscanf(line,"%s",header);
 printf ("%s\n",header);

 fgets(line,80,infile);
 sscanf(line,"%s",dummy);
 printf ("%s\n",dummy);
 
 fgets(line,80,infile);
 sscanf(line,"%i %i",&xres,&yres);
 printf("%i %i\n",xres,yres);

 fgets(line,80,infile);
 sscanf(line,"%i",&cres);
 printf("%i\n",cres);
 
 fclose(infile);

 /* Test header */
  
 if ( (!strncmp("P6",header,2) == 0) ) {
  if (BE_VERBOSE) fprintf (stderr,"Error: Not a PPM file.\n");
  return(the_image);
 }
 if ((!(xres == 640)) && (!(yres == 480))) {
  if (BE_VERBOSE) fprintf (stderr,"Error: Image not in 640x480 format.\n");
  return(the_image);
 }
 
 /* Create workspace */
 the_data=(char *)malloc(640*480*3);

 /* Load the file */
 infile=fopen(filename,"r");
 fgets(line,80,infile);
 fgets(line,80,infile);
 fgets(line,80,infile);
 fgets(line,80,infile);
 fread(the_data,640*480*3,1,infile);
 fclose(infile);
 
 /* Build the colortable  */
 for (i=0; i<640*480; i++) {
  j=3*i;
  r=the_data[j];
  g=the_data[j+1];
  b=the_data[j+2];
  cur_color=(r << 16) +
            (g << 8) +
            (b);
  if (num_colors==0) {
   the_colors[num_colors]=cur_color;
   num_colors++;
  } else {
   j=0;
   while ((j<num_colors) && (!(the_colors[j] == cur_color))) j++;
   if (j==num_colors) {
    if (num_colors==255) {
     if (BE_VERBOSE) fprintf (stderr,"Error: Image exceeds the 255 possible colors.\n");
     free(the_data);
     return(the_image);
    }
    the_colors[num_colors]=cur_color;
    num_colors++;
   } 
  }
 }
 if (BE_VERBOSE) fprintf (stderr,"Need %i colors so far.\n",num_colors);
 
 /* Decode the image  */
 for (i=0; i<640*480; i++) {
  j=3*i;
  cur_color=(the_data[j] << 16) |
            (the_data[j+1] << 8) |
            (the_data[j+2]);
  j=0;
  while ((j<num_colors) && (!(the_colors[j] == cur_color))) j++;
  the_image[i]=(j+1);
 }

 return (the_image);
}

/* Get a subsecond relative time (uptime) */

float GetRelativeTime()
{
 FILE *infile;
 float thetime;

 /* Get the uptime for FPS calculations */
 infile=fopen("/proc/uptime","r");
 fscanf(infile,"%f",&thetime);
 fclose(infile);

 return(thetime);
}

void init_blockcounters()
{
 int counter;
 int ix,iy;
 float lengthx,lengthy,stepx,stepy;
 
 if (BE_VERBOSE) fprintf (stderr,"====== Variable Setup ======\n\n");
 /* Loop over all screens */

 if (BE_VERBOSE) fprintf (stderr,"%i\n",sizeof(short));
 
 for (counter=0; counter<NUM_SCREENS; counter++) {

  /* Initialize variables */

  xstart[counter]=initial_value[counter][0]; 
  xsize[counter]=initial_value[counter][1];
  ystart[counter]=initial_value[counter][2]; 
  ysize[counter]=initial_value[counter][3];
 
  xblocksize[counter]=(int *)malloc(xsize[counter]*sizeof(int));
  yblocksize[counter]=(int *)malloc(ysize[counter]*sizeof(int));
  xtest[counter]=(int *)malloc(xsize[counter]*sizeof(int));

  stepx=(float)COLS/(float)xsize[counter];
  stepy=(float)ROWS/(float)ysize[counter];
  
  /* Setup X blocksizes */

  lengthx=0.0;
  for (ix=0; ix<xsize[counter]; ix++) {
   lengthx += stepx;
   xblocksize[counter][ix]=(int)(lengthx+0.5);
  }
  for (ix=(xsize[counter]-1); ix>0; ix--) {
   xblocksize[counter][ix]=xblocksize[counter][ix]-xblocksize[counter][ix-1];
  }
  
  /* Setup Y Blocksizes */

  lengthy=0.0;
  for (iy=0; iy<ysize[counter]; iy++) {
   lengthy += stepy;
   yblocksize[counter][iy]=(int)(lengthy+0.5);
  }
  for (iy=(ysize[counter]-1); iy>0; iy--) {
   yblocksize[counter][iy]=yblocksize[counter][iy]-yblocksize[counter][iy-1];
  }

  /* Print some info */
  if (BE_VERBOSE) {
   fprintf (stderr,"*** Layout for Screen #%i ***\n\n",counter);
   fprintf (stderr,"Blocks: (%i x %i)=%i\n",xsize[counter],ysize[counter],xsize[counter]*ysize[counter]);
   fprintf (stderr,"Blockwidths: ");
   for (ix=0; ix<xsize[counter]; ix++) {
    fprintf (stderr,"%i ",xblocksize[counter][ix]);
   }
   fprintf (stderr,"\n");
   fprintf (stderr,"Blockheights: ");
   for (iy=0; iy<ysize[counter]; iy++) {
    fprintf (stderr,"%i ",yblocksize[counter][iy]);
   }
   fprintf (stderr,"\n\n");
  } 
 }
}

/* Framegrabber service routine: process and output */

void gotframe(int x)
{
  int counter;
  int ix,iy,oy;
  short *in_root;
  int *test_root,*backup_test_root;
  int *xblock_root,*backup_xblock_root;
  int *yblock_root;
  char *out_root;
  char *trans_root;
  char *transbg_root;
  int test_size;
  int cur_xsize;
  int cur_ysize;
  int xloopsize;
  int nextrow;
  float TheCurrentTime;
  KeySym key;
  XEvent event; 

  /* Loop over all screens */

  for (counter=0; counter<NUM_SCREENS; counter++) {

   /* Setup input pointers */

   in_root=(short *)fbuf;
   in_root+=(xstart[counter]+(ystart[counter]-1)*COLS);

   /* Setup output pointer */

   out_root=(char *)shmimage[counter]->data;
   trans_root=(char *)tbuf[counter];
   transbg_root=(char *)bgbuf;

   /* Calc sizes */
   cur_xsize=xsize[counter];
   cur_ysize=ysize[counter];
   test_size=cur_xsize*sizeof(int);
   xloopsize=cur_xsize-1;
   nextrow=COLS-cur_xsize+1;
   
   /* Setup X arrays */
   
   backup_test_root = xtest[counter];
   backup_xblock_root = xblocksize[counter];

   /* Setup Y array */

   yblock_root=yblocksize[counter];
   
   /* Scan out input area, Y loop */
   
   for (iy=0; iy<cur_ysize; iy++) {

    /* Reset pointers */
    test_root = backup_test_root;;
    bzero(test_root,test_size);
    
    /* Check input line, X loop */

    for (ix=0; ix<xloopsize; ix++) {
     /* Test lookup */
     *test_root = colortest[counter][*in_root];
     /* Next output block test pointer */
     test_root++;
     /* Update pointers for input array */
     in_root++;
    }

    /* Draw/transfer lines based on previous test */

    for (oy=0; oy<*yblock_root; oy++) {

     /* Reset blocksize pointer */
     xblock_root = backup_xblock_root;

     /* Reset test array pointer */
     test_root = backup_test_root;

     /* Loop over one input line */
     for (ix=0; ix<cur_xsize; ix++) {
      if (*test_root) {
       /* Transfer background */
       /* bzero(out_root,*xblock_root); */
       memcpy(out_root,transbg_root,*xblock_root);
      } else {
       /* Transfer image */
       memcpy(out_root,trans_root,*xblock_root);
      } 
      out_root += *xblock_root;
      trans_root += *xblock_root;
      transbg_root += *xblock_root;
      /* Update test and blocksize array pointer */
      test_root++;
      xblock_root++;
     }
    } 

    /* Update y-blocksize pointer */
    yblock_root++;
    /* Go to next input row */
    in_root += nextrow;
   }

   // Draw screen onto display
   XShmPutImage(display_[counter], wtwin_[counter], 
		gc_[counter], shmimage[counter],
	   	0, 0, 0, 0, WINDOW_SIZE_X, WINDOW_SIZE_Y, False);
   XSync(display_[counter], 0);
 }
 
 /* Check keyboard/mouse input */
 if (XPending(display_[0])>0) {
  key=0;
  XNextEvent(display_[0], &event);
  if (event.type==KeyPress) {
   key=XLookupKeysym((void *)&event,0);
   /* Handle keyboard input */
   switch (key) {
    case XK_u:
     updateColortest();
     break;
    case XK_s:
     saveColortest(ColortestName);  
     break;
   }
  }
 }
   
 /* Every 100th frame, output FPS number */
 framecounter++;
 if (framecounter==100) {
  TheCurrentTime = GetRelativeTime();
  if (TheCurrentTime>LastTime) {
   if (BE_VERBOSE) fprintf (stderr,"Display Speed = %4.1f FPS\n",100.0/(TheCurrentTime-LastTime)); 
  }
  framecounter=0;
  LastTime=TheCurrentTime;
 }   
}

/* ---------- MAIN --------- */

int main(int argc, char **argv)
{
  int counter;
  struct meteor_geomet geo;
  struct meteor_frame_offset off;
  int i,c;


  if (BE_VERBOSE) {
   fprintf (stderr,"##########                                  ##########\n");
   fprintf (stderr,"##########      DBA VideoWall display       ##########\n");
   fprintf (stderr,"##########     A. Schiffler, 1997,1999      ##########\n");
   fprintf (stderr,"##########                                  ##########\n\n");
  }
  

  /* Setup signal handler */  
  sigact.sa_handler=gotframe;
  sigact.sa_flags=SA_RESTART;
  
  /* Load images and define colors from backimage */
  num_colors=0;
  for (counter=0; counter<NUM_SCREENS; counter++) {
   tbuf[counter] = LoadImage(imagenames[counter]);
  } 
  bgbuf = LoadImage(backimagename);

  /* Set up variables */
  init_blockcounters();
  initColortest();

  /* Load test array */
  loadColortest(ColortestName);

  /* Open framegrabber */
  if ( argc > 1 )
    DevName = *++argv;
  else
    DevName = DefaultName;
  if ((i = open(DevName, O_RDONLY)) < 0) {
    perror("open failed");
    printf("device %s\n", DevName);
    exit(1);
  }

  /* Fire up the screens */
  setupDisplay();

  /* Setup meteor */
  geo.rows = ROWS;
  geo.columns = COLS;
  geo.frames = 1;
  geo.oformat = METEOR_GEO_RGB16;

  if (ioctl(i, METEORSETGEO, &geo) < 0) {
    perror("ioctl SetGeometry failed");
    exit(1);
  }
  if (ioctl(i, METEORGFROFF, &off) < 0) {
    perror("ioctl GetFrameOffset failed");
    exit(1);
  }

  c = METEOR_FMT_PAL;
  if (ioctl(i, METEORSFMT, &c) < 0) {
    perror("ioctl SetFormat failed");
    exit(1);
  }

  c = METEOR_INPUT_DEV0;
  if (ioctl(i, METEORSINPUT, &c) < 0) {
    perror("ioctl Setinput failed");
    exit(1);
  }

  mmbuf=(char *)mmap((caddr_t)0, RGB_SIZE, PROT_READ,MAP_FILE|MAP_SHARED, i, (off_t)0);
  if ( !mmbuf ) {
    perror("mmap failed");
    exit(1);
  }
  fbuf = mmbuf + off.frame_offset[0];

  if ( sigaction(SIGUSR2, &sigact, NULL) ) {
    perror("sigaction failed");
    exit(1);
  }

  c = SIGUSR2;
  if (ioctl(i, METEORSSIGNAL, &c) < 0) {
    perror("ioctl SetSignal failed");
    exit(1);
  }

  /* Setup (reset) FPS variables */
  LastTime=GetRelativeTime();
  framecounter=0;
  
  /* Continuous frame capture */
  c = METEOR_CAP_CONTINOUS ;
  if (ioctl(i, METEORCAPTUR, &c)) {
    perror("ioctl CaptContinuous failed");
    exit(1);
  }
  while(1)
    pause();
}

static void setupDisplay() {
  int counter;
  XGCValues gc_values;
  XVisualInfo desired_visual_template;
  XVisualInfo *visual_info_list; 
  int no_visual_matched;
  int depth;
  int screen;
  int pixelNum;
  XWindowAttributes win_attributes;
  char displaystring[10];
  int displaywin;
  
  if (BE_VERBOSE) fprintf (stderr,"===== Setting up screens =======\n");

  for (counter=0; counter<NUM_SCREENS; counter++) {

   if (SINGLE_SCREEN) {
    displaywin=0;
   } else {
    displaywin=counter;
   } 
   sprintf (displaystring,":0.%i",displaywin);

   display_[counter] = XOpenDisplay(displaystring);
   wtwin_[counter] = XCreateSimpleWindow(display_[counter], 
				DefaultRootWindow(display_[counter]),
			        0, 0, WINDOW_SIZE_X, WINDOW_SIZE_Y, 2,
			        BlackPixel(display_[counter], 
				 DefaultScreen(display_[counter])),
			        BlackPixel(display_[counter], 
				 DefaultScreen(display_[counter])));  
   XMapWindow(display_[counter], wtwin_[counter]);
   XSelectInput(display_[counter], wtwin_[counter],KeyPressMask | KeyReleaseMask);
   gc_values.graphics_exposures = False;
   gc_[counter] = XCreateGC(display_[counter], 
			wtwin_[counter], GCGraphicsExposures, &gc_values);  

   //SHARED MEMORY PORTION

   screen = DefaultScreen(display_[counter]);
   XGetWindowAttributes(display_[counter], wtwin_[counter], &win_attributes);
   depth = win_attributes.depth;
  
   shminfo[counter] = (XShmSegmentInfo*) malloc (sizeof(XShmSegmentInfo));
   shmimage[counter] = XShmCreateImage(display_[counter], DefaultVisual(display_[counter], screen),
			     depth, ZPixmap, NULL, shminfo[counter], WINDOW_SIZE_X,
			     WINDOW_SIZE_Y);
   shminfo[counter]->shmid = shmget(ftok(DevName, 'v'),
			  shmimage[counter]->bytes_per_line * shmimage[counter]->height,
			  IPC_CREAT | 0777);
   shminfo[counter]->shmaddr = (char *) shmat(shminfo[counter]->shmid, (void *) 0, 0);
   shmimage[counter]->data = shminfo[counter]->shmaddr;
  
   XShmAttach(display_[counter], shminfo[counter]);

   // Set palette

   desired_visual_template.screen = screen;
   desired_visual_template.class = PseudoColor;
   visual_info_list = XGetVisualInfo(display_[counter], 
				    VisualScreenMask | VisualClassMask,
				    &desired_visual_template,
				    &no_visual_matched);
   if (no_visual_matched == 0) {
    if (BE_VERBOSE) fprintf (stderr,"Error: No visual matched - set screen to 256 colors.");
    exit(0);
   }

   wtcolormap[counter] = XCreateColormap(display_[counter], 
				DefaultRootWindow(display_[counter]),
			        visual_info_list[0].visual, AllocAll);

   /* Colors */
   for (pixelNum=0;pixelNum<maxcols;pixelNum++) {
     palette[counter][pixelNum].pixel=pixelNum;
     palette[counter][pixelNum].red=((the_colors[pixelNum-1] >> 16 & 255))*256;
     palette[counter][pixelNum].green=((the_colors[pixelNum-1] >> 8) & 255)*256;
     palette[counter][pixelNum].blue=(the_colors[pixelNum-1] & 255)*256;
     palette[counter][pixelNum].flags=DoRed|DoGreen|DoBlue;
   }

   XStoreColors(display_[counter], wtcolormap[counter], palette[counter], maxcols);
   XSetWindowColormap(display_[counter], wtwin_[counter], wtcolormap[counter]);
   
   XInstallColormap(display_[counter],wtcolormap[counter]);
 
  }
}
